@@ -25,6 +25,8 @@ from .utils import (
     convert_to_seconds,
     set_sigterm,
     get_speaker,
+    save_search,
+    read_search,
 )
 
 
@@ -265,6 +267,22 @@ def print_info(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
+@zero_parameters
+def track(speaker, action, args, soco_function, use_local_speaker_list):
+    if speaker.is_playing_line_in:
+        print("Playing from Line In")
+    else:
+        output = speaker.get_current_track_info()
+        if output["duration"] == "0:00:00":
+            # Assume it's a radio station
+            print("Playing {}".format(output["title"]))
+        else:
+            for item in sorted(output):
+                if item not in ["metadata", "uri", "album_art"]:
+                    print("  {}: {}".format(item.capitalize(), output[item]))
+    return True
+
+
 @zero_or_one_parameter
 def playback_mode(speaker, action, args, soco_function, use_local_speaker_list):
     np = len(args)
@@ -359,8 +377,7 @@ def transport_state(speaker, action, args, soco_function, use_local_speaker_list
 
 
 def play_favourite_core(speaker, favourite):
-    """Core of the play_favourite action, but doesn't exit on failure
-    """
+    """Core of the play_favourite action, but doesn't exit on failure"""
     fs = speaker.music_library.get_sonos_favorites(complete_result=True)
     the_fav = None
     # Strict match
@@ -416,7 +433,7 @@ def play_favourite(speaker, action, args, soco_function, use_local_speaker_list)
         return True
 
 
-@one_parameter
+@one_or_two_parameters
 def add_favourite_to_queue(
     speaker, action, args, soco_function, use_local_speaker_list
 ):
@@ -438,9 +455,25 @@ def add_favourite_to_queue(
                 the_fav = f
                 break
     if the_fav:
+        position = 0
+        if len(args) == 2:
+            if args[1].lower() not in ["play_next", "next"]:
+                error_and_exit(
+                    "If supplied, second parameter must be 'play_next' or 'next'"
+                )
+            if (
+                speaker.get_current_transport_info()["current_transport_state"]
+                == "PLAYING"
+            ):
+                offset = 1
+            else:
+                offset = 0
+            position = (
+                int(speaker.get_current_track_info()["playlist_position"]) + offset
+            )
         try:
             # Print the queue position and return
-            print(speaker.add_to_queue(the_fav))
+            print(speaker.add_to_queue(the_fav, position=position))
             return True
         except Exception as e:
             error_and_exit("{}".format(str(e)))
@@ -452,7 +485,9 @@ def add_favourite_to_queue(
 @one_parameter
 def play_favourite_radio(speaker, action, args, soco_function, use_local_speaker_list):
     favourite = args[0]
-    fs = speaker.music_library.get_favorite_radio_stations()
+    preset = 0
+    limit = 99
+    fs = speaker.music_library.get_favorite_radio_stations(preset, limit)
     the_fav = None
     # Strict match
     for f in fs:
@@ -469,28 +504,29 @@ def play_favourite_radio(speaker, action, args, soco_function, use_local_speaker
                 the_fav = f
                 break
     if the_fav:
-        # play_uri works for some favourites
-        try:
-            uri = the_fav.get_uri()
-            metadata = the_fav.resource_meta_data
-            "Trying 'play_uri()': URI={}, Metadata={}".format(uri, metadata)
-            speaker.play_uri(uri=uri, meta=metadata)
-            return True
-        except Exception as e:
-            e1 = e
-            pass
-        # Other favourites will be added to the queue, then played
-        try:
-            # Add to the end of the current queue and play
-            index = speaker.add_to_queue(the_fav, as_next=True)
-            logging.info("Used'add_to_queue() [at {}], then play'".format(index))
-            speaker.play_from_queue(index, start=True)
-            return True
-        except Exception as e2:
-            error_and_exit("1: {} | 2:{}".format(str(e1), str(e2)))
-            return False
-    error_and_exit("Favourite '{}' not found".format(args[0]))
-    return False
+        uri = the_fav.get_uri()
+        meta_template = """
+        <DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/"
+            xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
+            xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
+            xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">
+            <item id="R:0/0/0" parentID="R:0/0" restricted="true">
+                <dc:title>{title}</dc:title>
+                <upnp:class>object.item.audioItem.audioBroadcast</upnp:class>
+                <desc id="cdudn" nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
+                    {service}
+                </desc>
+            </item>
+        </DIDL-Lite>' """
+        tunein_service = "SA_RINCON65031_"
+        uri = uri.replace("&", "&amp;")
+        metadata = meta_template.format(title=the_fav.title, service=tunein_service)
+        logging.info("Trying 'play_uri()': URI={}, Metadata={}".format(uri, metadata))
+        speaker.play_uri(uri=uri, meta=metadata)
+        return True
+    else:
+        error_and_exit("Favourite '{}' not found".format(args[0]))
+        return False
 
 
 @one_or_two_parameters
@@ -498,7 +534,9 @@ def play_uri(speaker, action, args, soco_function, use_local_speaker_list):
     force_radio = True if args[0][:4].lower() == "http" else False
     if len(args) == 2:
         speaker.play_uri(
-            args[0], title=args[1], force_radio=force_radio,
+            args[0],
+            title=args[1],
+            force_radio=force_radio,
         )
     else:
         speaker.play_uri(args[0], force_radio=force_radio)
@@ -645,7 +683,8 @@ def remove_from_queue(speaker, action, args, soco_function, use_local_speaker_li
     # Catch any non-integer input values
     except ValueError:
         parameter_type_error(
-            action, "integer, or comma-separated integers without spaces (e.g., 3,7,4)",
+            action,
+            "integer, or comma-separated integers without spaces (e.g., 3,7,4)",
         )
         return False
     # Catch any out-of-range values
@@ -726,7 +765,7 @@ def seek(speaker, action, args, soco_function, use_local_speaker_list):
     return True
 
 
-@one_parameter
+@one_or_two_parameters
 def playlist_operations(speaker, action, args, soco_function, use_local_speaker_list):
     name = args[0]
     if soco_function == "create_sonos_playlist":
@@ -737,9 +776,28 @@ def playlist_operations(speaker, action, args, soco_function, use_local_speaker_
         return True
     playlist = get_playlist(speaker, name)
     if playlist is not None:
-        result = getattr(speaker, soco_function)(playlist)
-        if soco_function in ["add_to_queue"]:
+        if soco_function == "add_to_queue":
+            position = 0
+            if len(args) == 2:
+                if args[1].lower() not in ["next", "play_next"]:
+                    error_and_exit(
+                        "If supplied, second parameter must be 'play_next' or 'next'"
+                    )
+                    return False
+                if (
+                    speaker.get_current_transport_info()["current_transport_state"]
+                    == "PLAYING"
+                ):
+                    offset = 1
+                else:
+                    offset = 0
+                position = (
+                    int(speaker.get_current_track_info()["playlist_position"]) + offset
+                )
+            result = speaker.add_to_queue(playlist, position=position)
             print(result)
+        else:
+            result = getattr(speaker, soco_function)(playlist)
     else:
         error_and_exit("Playlist '{}' not found".format(args[0]))
         return False
@@ -891,7 +949,8 @@ def groups(speaker, action, args, soco_function, use_local_speaker_list):
             print("[{}] : ".format(group.short_label), end="")
             for member in group.members:
                 print(
-                    "{} ({}) ".format(member.player_name, member.ip_address), end="",
+                    "{} ({}) ".format(member.player_name, member.ip_address),
+                    end="",
                 )
             print()
     return True
@@ -943,6 +1002,7 @@ def list_alarms(speaker, action, args, soco_function, use_local_speaker_list):
     ]
     print()
     print(tabulate.tabulate(sorted(details), headers, numalign="center"))
+    print()
     return True
 
 
@@ -1143,6 +1203,7 @@ def search_albums(speaker, action, args, soco_function, use_local_speaker_list):
         print_list_header("Sonos Music Library Album Search:", name)
         print_albums(albums)
         print()
+        save_search(albums)
     return True
 
 
@@ -1158,6 +1219,7 @@ def search_tracks(speaker, action, args, soco_function, use_local_speaker_list):
         print_list_header("Sonos Music Library Track Search:", name)
         print_tracks(tracks)
         print()
+        save_search(tracks)
     return True
 
 
@@ -1188,46 +1250,57 @@ def tracks_in_album(speaker, action, args, soco_function, use_local_speaker_list
     return True
 
 
-@one_parameter
+def queue_item_core(speaker, action, args, type):
+    name = args[0]
+    items = speaker.music_library.get_music_library_information(
+        type, search_term=name, complete_result=True
+    )
+    if len(items):
+        position = 0
+        if len(args) == 2:
+            if args[1].lower() in ["play_next", "next"]:
+                # Check if currently playing from the queue
+                if (
+                    speaker.get_current_transport_info()["current_transport_state"]
+                    == "PLAYING"
+                    and speaker.get_current_track_info()["position"]
+                    != "NOT_IMPLEMENTED"
+                ):
+                    offset = 1
+                else:
+                    offset = 0
+                position = (
+                    int(speaker.get_current_track_info()["playlist_position"]) + offset
+                )
+            else:
+                error_and_exit(
+                    "If supplied, second parameter for '{}' must be 'next/play_next'".format(
+                        action
+                    )
+                )
+                return False
+        # Select a random entry from the list
+        item = items[randint(0, len(items) - 1)]
+        print(speaker.add_to_queue(item, position=position))
+        return True
+    else:
+        error_and_exit("'{}' not found".format(name))
+        return False
+
+
+@one_or_two_parameters
 def queue_album(speaker, action, args, soco_function, use_local_speaker_list):
-    """Add an album to the queue. If there are multiple matches, a random
-    match will be selected.
-    :returns: The position in the queue of the first track in the album
-    """
-    name = args[0]
-    albums = speaker.music_library.get_music_library_information(
-        "albums", search_term=name, complete_result=True
-    )
-    if len(albums):
-        album = albums[randint(0, len(albums) - 1)]
-        print(speaker.add_to_queue(album))
-        return True
-    else:
-        error_and_exit("Album '{}' not found".format(name))
+    return queue_item_core(speaker, action, args, "albums")
 
 
-@one_parameter
+@one_or_two_parameters
 def queue_track(speaker, action, args, soco_function, use_local_speaker_list):
-    """Add a track to the queue. If there are multiple matches, a random match
-    will be selected.
-    :returns: The position in the queue of the track
-    """
-    name = args[0]
-    tracks = speaker.music_library.get_music_library_information(
-        "tracks", search_term=name, complete_result=True
-    )
-    if len(tracks):
-        track = tracks[randint(0, len(tracks) - 1)]
-        print(speaker.add_to_queue(track))
-        return True
-    else:
-        error_and_exit("Track '{}' not found".format(name))
+    return queue_item_core(speaker, action, args, "tracks")
 
 
 @one_or_more_parameters
 def if_stopped_or_playing(speaker, action, args, soco_function, use_local_speaker_list):
-    """Perform the action only if the speaker is currently in the desired playback state
-    """
+    """Perform the action only if the speaker is currently in the desired playback state"""
     state = speaker.get_current_transport_info()["current_transport_state"]
     logging.info(
         "Condition: '{}': Speaker '{}' is in state '{}'".format(
@@ -1268,7 +1341,13 @@ def cue_favourite(speaker, action, args, soco_function, use_local_speaker_list):
     if not speaker.group.mute:
         speaker.group.mute = True
         unmute_group = True
-    result, msg = play_favourite_core(speaker, args[0])
+    if action in ["cfrs", "cue_favourite_radio_station"]:
+        result = play_favourite_radio(
+            speaker, action, args, soco_function, use_local_speaker_list
+        )
+        msg = ""
+    else:
+        result, msg = play_favourite_core(speaker, args[0])
     speaker.stop()
     if unmute:
         speaker.mute = False
@@ -1282,8 +1361,7 @@ def cue_favourite(speaker, action, args, soco_function, use_local_speaker_list):
 
 @one_parameter
 def transfer_playback(speaker, action, args, soco_function, use_local_speaker_list):
-    """Transfer playback from one speaker to another, by grouping and ungrouping.
-    """
+    """Transfer playback from one speaker to another, by grouping and ungrouping."""
     if not speaker.is_coordinator:
         error_and_exit("Speaker '{}' is not a coordinator".format(speaker.player_name))
         return False
@@ -1297,11 +1375,102 @@ def transfer_playback(speaker, action, args, soco_function, use_local_speaker_li
         return False
 
 
+@zero_parameters
+def queue_position(speaker, action, args, soco_function, use_local_speaker_list):
+    print(speaker.get_current_track_info()["playlist_position"])
+    return True
+
+
+@zero_parameters
+def last_search(speaker, action, args, soco_function, use_local_speaker_list):
+    items = read_search()
+    if items:
+        if len(items):
+            print()
+            print_list_header(
+                "Sonos Music Library: Saved {} Search".format(
+                    items.search_type.capitalize()
+                ),
+                "",
+            )
+            if items.search_type == "albums":
+                print_albums(items)
+            elif items.search_type == "tracks":
+                print_tracks(items)
+            print()
+    else:
+        error_and_exit("No saved search")
+        return False
+    return True
+
+
+@one_or_two_parameters
+def queue_search_result_number(
+    speaker, action, args, soco_function, use_local_speaker_list
+):
+    try:
+        saved_search_number = int(args[0])
+    except ValueError:
+        parameter_type_error(
+            action, "An integer index from the previous search results"
+        )
+        return False
+    items = read_search()
+    if not items:
+        error_and_exit("No saved search")
+        return False
+    logging.info("Loaded saved search")
+    position = 0
+    if len(args) == 2:
+        if args[1].lower() in ["play_next", "next"]:
+            # Check if currently playing from the queue
+            # If so, add at the next track position
+            if (
+                speaker.get_current_transport_info()["current_transport_state"]
+                == "PLAYING"
+                and speaker.get_current_track_info()["position"] != "NOT_IMPLEMENTED"
+            ):
+                logging.info("Currently playing from queue; add as next track")
+                offset = 1
+            # Otherwise use the current position
+            else:
+                logging.info("Not currently playing; add at current queue position")
+                offset = 0
+            position = (
+                int(speaker.get_current_track_info()["playlist_position"]) + offset
+            )
+        else:
+            error_and_exit(
+                "If supplied, second parameter for '{}' must be 'next/play_next'".format(
+                    action
+                )
+            )
+            return False
+    # Select the item number from the saved search
+    if 1 <= saved_search_number <= len(items):
+        item = items[saved_search_number - 1]
+        print(speaker.add_to_queue(item, position=position))
+        return True
+    else:
+        error_and_exit("Item search index must be between 1 and {}".format(len(items)))
+        return False
+
+
+def cue_favourite_radio_station(
+    speaker, action, args, soco_function, use_local_speaker_list
+):
+    return cue_favourite(speaker, action, args, soco_function, use_local_speaker_list)
+
+
 def process_action(speaker, action, args, use_local_speaker_list):
     sonos_function = actions.get(action, None)
     if sonos_function:
         return sonos_function.processing_function(
-            speaker, action, args, sonos_function.soco_function, use_local_speaker_list,
+            speaker,
+            action,
+            args,
+            sonos_function.soco_function,
+            use_local_speaker_list,
         )
     else:
         return False
@@ -1309,7 +1478,12 @@ def process_action(speaker, action, args, use_local_speaker_list):
 
 # Type for holding action processing functions
 SonosFunction = namedtuple(
-    "SonosFunction", ["processing_function", "soco_function",], rename=False,
+    "SonosFunction",
+    [
+        "processing_function",
+        "soco_function",
+    ],
+    rename=False,
 )
 
 # Actions and associated processing functions
@@ -1359,7 +1533,7 @@ actions = {
     "group_relative_volume": SonosFunction(relative_volume, "group_relative_volume"),
     "group_rel_vol": SonosFunction(relative_volume, "group_relative_volume"),
     "grv": SonosFunction(relative_volume, "group_relative_volume"),
-    "track": SonosFunction(print_info, "get_current_track_info"),
+    "track": SonosFunction(track, ""),
     "play_mode": SonosFunction(playback_mode, "play_mode"),
     "mode": SonosFunction(playback_mode, "play_mode"),
     "playback_state": SonosFunction(transport_state, "get_current_transport_info"),
@@ -1440,6 +1614,8 @@ actions = {
     "favourite_radio_stations": SonosFunction(
         list_numbered_things, "get_favorite_radio_stations"
     ),
+    "frs": SonosFunction(list_numbered_things, "get_favorite_radio_stations"),
+    "lfrs": SonosFunction(list_numbered_things, "get_favorite_radio_stations"),
     "play_favourite_radio_station": SonosFunction(play_favourite_radio, "play_uri"),
     "play_favorite_radio_station": SonosFunction(play_favourite_radio, "play_uri"),
     "pfrs": SonosFunction(play_favourite_radio, "play_uri"),
@@ -1498,4 +1674,14 @@ actions = {
     "rctfq": SonosFunction(remove_current_track_from_queue, ""),
     "remove_last_track_from_queue": SonosFunction(remove_last_track_from_queue, ""),
     "rltfq": SonosFunction(remove_last_track_from_queue, ""),
+    "queue_position": SonosFunction(queue_position, ""),
+    "qp": SonosFunction(queue_position, ""),
+    "last_search": SonosFunction(last_search, ""),
+    "ls": SonosFunction(last_search, ""),
+    "queue_search_result_number": SonosFunction(queue_search_result_number, ""),
+    "queue_search_number": SonosFunction(queue_search_result_number, ""),
+    "qsn": SonosFunction(queue_search_result_number, ""),
+    "cue_favourite_radio_station": SonosFunction(cue_favourite_radio_station, ""),
+    "cue_favorite_radio_station": SonosFunction(cue_favourite_radio_station, ""),
+    "cfrs": SonosFunction(cue_favourite_radio_station, ""),
 }
